import torch
import torch.nn.functional as F

def calculate_returns(rewards, discount_factor, normalize = True, device = 'cpu'):
    returns = []
    R = 0

    for r in reversed(rewards):
        R = r + discount_factor * R
        returns.insert(0, R)

    returns = torch.tensor(returns).to(device=device)
    if normalize:
        returns = (returns - returns.mean()) / returns.std()

    return returns


def update_policy(returns, log_prob_actions, values, optimizer):
    returns = returns.detach()
    
    policy_loss = - (returns * log_prob_actions).sum()

    criterion = nn.MSELoss()
    value_loss = torch.sqrt(criterion(returns, values))
    

    optimizer.zero_grad()

    value_loss.backward()

    optimizer.step()

    return value_loss.item()


def evaluate(env, policy):
    policy.eval()

    done = False
    episode_reward = 0
    state = env.reset()

    while not done:
        state = torch.FloatTensor(state)
        with torch.no_grad():
            action_pred, _ = policy(state)
            action_prob = F.softmax(action_pred, dim=-1)
        action = torch.argmax(action_prob, dim = -1)
        state, reward, done, _ = env.step(action.item())
        episode_reward += reward
    return episode_reward

def train(env, q_model, importance,  optimizer, discount, device = 'cpu'):
  log_prob_actions = []
  q_values = []
  reward_values = []
  done_values = False
  episode_reward = 0
  
  current_observations, next_action, next_observations, reward, done = env.reset()
  step = 0
  while not done:
    
    action,q = q_model(current_observations)
    
    action_prob = F.softmax(action, dim=-1)
    action_dist = distributions.Categorical(action_prob)
  
    log_prob_action = action_dist.log_prob(action[next_action])
      
    q_values.append(q[next_action])
    reward_values.append(reward.item())

    log_prob_actions.append(log_prob_action)

    episode_reward += reward
    current_observations, next_action, next_observations, reward, done = env.step()
    step += 1
  log_prob_actions = torch.stack(log_prob_actions)
  values = torch.stack(q_values)
  returns = calculate_returns(reward_values, discount, device=device)
  value_loss = update_policy(returns, log_prob_actions, values, optimizer)
    
  return value_loss, episode_reward